/*
 * tqueue.c --- task queue handling for Linux.
 *
 * This routine merely draws in the static portion of the task queue
 * inline functions.  Look in tqueue.h for the relevant functions.
 */

#define INCLUDE_INLINE_FUNCS

#include <linux/tqueue.h>
